--
-- GameExtensionUtils
--
-- @author:    	Xentro (Marcus@Xentro.se)
-- @website:	www.Xentro.se
-- @history:	
-- 

-- Debuging --
GameExtension.visualDebug = {};

-- Print message to log
function log(mode, message)
	local debugLevel = -1; -- Forced debug
	local t = "DEBUG";
	
	if mode ~= nil and type(mode) == "string" and GameExtension[mode] ~= nil then
		debugLevel = GameExtension[mode];
		t = mode;
	end;
	
	if debugLevel <= GameExtension.MESSAGE_MODE then
		print("  GameExtension (v" .. GameExtension.version .. ") - " .. t .. ": " .. tostring(message));
	end;
end;

-- Print table to log
function logTable(table, depth, name)
	DebugUtil.printTableRecursively(table, Utils.getNoNil(name, ""), 0, Utils.getNoNil(depth, 3));
end;

-- Show message on screen real time
-- 	 showMessage("testing", "now what?");
function showMessage(name, message)
	addMessage(name, message, nil, true); -- replace value 
end;

-- Show message on screen, Usage examples
-- (fetch variable value)
-- 	  addMessage("moneyUnit", g_currentMission.missionInfo);
-- (fetch variable value with custom name)
-- 	  addMessage("Our money unit:", g_currentMission.missionInfo, "moneyUnit");
function addMessage(name, parent, variable, replace)
	if GameExtension.MESSAGE_MODE == GameExtension.DEBUG then
		if GameExtension.visualDebug[name] == nil then
			GameExtension.visualDebug[name] = {variable = Utils.getNoNil(variable, name), parent = parent};
		else
			if replace ~= nil and replace then
				GameExtension.visualDebug[name].parent = parent;
			end;
		end;
	end;
end;

function removeMessage(name)
	GameExtension.visualDebug[name] = nil;
end;



-- Modules --

function GameExtension.addModule(moduleName, isSpec, state, settings, sortedKeys)
	if GameExtension[moduleName] == nil then
		GameExtension[moduleName] = Utils.getNoNil(state, true); -- State of module
		GameExtension.modules[moduleName] = Utils.getNoNil(settings, {});
		GameExtension.modules[moduleName].isSpec = isSpec;
		GameExtension.modules[moduleName].sortedKeys = Utils.getNoNil(sortedKeys, {});
	end;
end;

--[[ call in load function, all mods are loaded by then and its safe to deactivate modules.
local mod = _G["FS17_GameExtension"];
if mod ~= nil then
	mod.GameExtension.deactivateModule("CRUISE_CONTROL", g_currentModName); 
end; 
]]
function GameExtension.deactivateModule(moduleName, name, force)
	moduleName = moduleName:upper();
	
	if GameExtension.getSetting(moduleName) and name ~= nil then
		if (GameExtension.getSetting(moduleName, "ALLOW_DEACTIVATE") or force ~= nil and force) then
			GameExtension.addBlackListItem(moduleName);
			GameExtension.setSetting(moduleName, nil, false, true);
			log("NOTICE", "Module " .. moduleName .. " are being deactivated by the mod " .. name);
		end;
	end;
end;



-- Specializations --

function GameExtension.addSpecialization(name, spec)
	-- log("DEBUG", "Specialization " .. name .. " is added.");
	
	table.insert(GameExtension.specializations, spec);
end;



-- Settings --

function GameExtension.getSetting(moduleName, setting, getSettings)
	if GameExtension[moduleName] ~= nil then
		if setting == nil and not getSettings then
			return GameExtension[moduleName];
		elseif getSettings ~= nil and getSettings then
			return GameExtension.modules[moduleName];
		elseif setting ~= nil and GameExtension.modules[moduleName][setting] ~= nil then
			return GameExtension.modules[moduleName][setting].value;
		end;
	else
		return nil;
	end;
end;

function GameExtension.setSetting(moduleName, setting, newValue, noEventSend, loadDefault)
	if GameExtension[moduleName] ~= nil and newValue ~= nil then
		local formatType = "bool";
		
		if setting ~= nil then
			if GameExtension.modules[moduleName][setting] ~= nil or loadDefault ~= nil and loadDefault then
				if not loadDefault then
					-- if not GameExtension.modules[moduleName][setting].isLocked then -- We shouldn't need this.
						if type(GameExtension.modules[moduleName][setting].value) ~= type(newValue) then
							log("NOTICE", "Setting " .. setting .. " is using the wrong type " .. type(newValue) .. " when it should be " .. type(GameExtension.modules[moduleName][setting].value) .. " type.");
							newValue = GameExtension.modules[moduleName][setting].value;
						end;
						
						GameExtension.modules[moduleName][setting].value = newValue;
						
						-- Dont call if module are deactivated
						if GameExtension[moduleName] then
							-- module files have been loaded now
							local v = GameExtension.modules[moduleName];
							if v[setting].func ~= nil then
								v.class[v[setting].func](g_currentMission.GameExtension, newValue);
							end;
						end;
					-- end;
				else
					GameExtension.modules[moduleName][setting] = newValue;
					table.insert(GameExtension.modules[moduleName].sortedKeys, setting);
				end;
				
				formatType = GameExtension.modules[moduleName][setting].type;
				
				-- log("DEBUG", setting .. " - " .. tostring(GameExtension.modules[moduleName][setting].value));
			end;
		else
			-- We should check ALLOW_DEACTIVATE here
			GameExtension[moduleName] = newValue;
		end;
		
		-- We make sure to not send events when we setup the settings table
		if g_gameExtensionGUI ~= nil then
			local sendEvent = true; -- module state
			
			if setting ~= nil then
				sendEvent = GameExtension.modules[moduleName][setting].allowEvent;
			end;
			
			if sendEvent then
				SetSettingEvent.sendEvent(moduleName, setting, formatType, newValue, noEventSend);
			else
				-- Save client settings when they are changed but only once everything is loaded and ready to go.
				if g_currentMission.GameExtension ~= nil and g_currentMission.GameExtension.firstTimeRun then
					if g_currentMission:getIsClient() and g_dedicatedServerInfo == nil then
						GameExtension.saveFile(GameExtension.clientFilename, false);
					end;
				end;
			end;
		end;
	end;
end;

function GameExtension.lockSetting(moduleName, setting, lock, force)
	if GameExtension.modules[moduleName] ~= nil and GameExtension.modules[moduleName][setting] ~= nil then
		if not force then
			GameExtension.modules[moduleName][setting].isLocked = lock;			-- This is set by GameExtension
		else
			GameExtension.modules[moduleName][setting].isLockedByForce = lock;	-- Modders need to use this!
		end;
	
		-- Maybe we should be doing an event here?
	end;
end;

function GameExtension.showSettingForMod(modName, moduleName, setting, state)
	if GameExtension.tempShowSetting == nil then
		GameExtension.tempShowSetting = {};
	end;
	
	if GameExtension.modules[moduleName] ~= nil and GameExtension.modules[moduleName][setting] ~= nil then
		table.insert(GameExtension.tempShowSetting, {modName = modName, moduleName = moduleName, setting = setting, state = Utils.getNoNil(state, true)});
	end;
end;



-- Black List (For Settings) --

function GameExtension.getBlackListItem(name)
	if GameExtension.settingBlackList[name] ~= nil then
		return GameExtension.settingBlackList[name];
	end;
	
	return false;
end;

function GameExtension.addBlackListItem(name)
	if GameExtension.settingBlackList == nil then
		GameExtension.settingBlackList = {};
	end;
	
	if GameExtension.settingBlackList[name] == nil then
		GameExtension.settingBlackList[name] = true;
	end;
end;



-- Setup, Load, Save Settings --

function GameExtension.saveFile(filename, isServer)
	local filename = Utils.getFilename(filename, getUserProfileAppPath());
	local xmlFile = createXMLFile("savingGameExtension", filename, "GameExtension");
	local txt = "server";
	local i = 0;
	
	if not isServer then txt = "client"; end;
	
	log("DEBUG", "Starting to save " .. txt .. " settings");
	
	for moduleName, settings in pairs(GameExtension.modules) do
		if not GameExtension.getBlackListItem(moduleName) then
			local key, valid = string.format("GameExtension.module(%d)", i), false;
			
			local subI = 0;
			for name, v in pairs(settings) do
				local subKey = string.format(key .. ".setting(%d)", subI);
				
				if not GameExtension.getBlackListItem(name) then
					if (isServer and v.allowEvent or not isServer and not v.allowEvent) then
						setXMLString(xmlFile, subKey .. "#name", name);
						
						if v.type == "string" then
							setXMLString(xmlFile, subKey .. "#value", v.value);
						elseif v.type == "float" then
							setXMLFloat(xmlFile, subKey .. "#value", v.value);
						elseif v.type == "int" then
							setXMLInt(xmlFile, subKey .. "#value", v.value);
						elseif v.type == "bool" then
							setXMLBool(xmlFile, subKey .. "#value", v.value);
						end;
						
						setXMLString(xmlFile, subKey .. "#type", v.type);
						
						valid = true;
						subI = subI + 1;
						
						-- log("DEBUG", name .. " - " .. tostring(v.value));
					end;
				end;
			end;
			
			-- Only save for servers! client dont need to know the state of module here.
			if isServer then
				if GameExtension.getSetting(moduleName, "ALLOW_DEACTIVATE") then
					setXMLBool(xmlFile, key .. "#active", GameExtension.getSetting(moduleName));
					valid = true;
				end;
			end;
			
			if valid then
				setXMLString(xmlFile, key .. "#name", moduleName);
				i = i + 1;
			end;
		end;
	end;
	
	if i > 0 then -- save XML if we got something to save..
		if not saveXMLFile(xmlFile) then
			log("ERROR", "Cant save settings - xmlFile: " .. xmlFile .. ", File: " .. filename);
		else
			log("DEBUG", "Saved Settings - xmlFile ID: " .. xmlFile .. " - File Exists: " .. tostring(fileExists(filename)) .. " - File Path: " .. filename);
		end;
	end;
	
	delete(xmlFile);
end;

-- Call 1, setup settings - Call 2, load saved settings
function GameExtension.loadSettingXML(xmlPath, savedGame, isServer)
	local txt = "Default";
	local mes = "Loading " .. txt .. " Settings - File Exists: " .. tostring(fileExists(xmlPath)) .. " - File Path: " .. tostring(xmlPath);
	
	if savedGame then
		txt = "Saved";
		log("DEBUG", mes); -- wont call on loading of default xml
	end;
	
	if fileExists(xmlPath) then
		local xmlFile 			= loadXMLFile("GameExtension", xmlPath);
		local isMultiplayer 	= false;
		local deactivateOnMP 	= savedGame;
		
		if not savedGame then
			level = Utils.getNoNil(getXMLInt(xmlFile, "GameExtension#Debug"), GameExtension.NOTICE);
			GameExtension.MESSAGE_MODE = level;
			
			if GameExtension.MESSAGE_MODE == GameExtension.DEBUG then
				log("NOTICE", "Debug mode is activated.");
				log("DEBUG", mes); -- print in log that default xml is loaded
			end;
		else
			isMultiplayer = g_currentMission.missionDynamicInfo.isMultiplayer;
		end;
		
		local i = 0;
		while true do
			local key = string.format("GameExtension.module(%d)", i);
			if not hasXMLProperty(xmlFile, key) then break; end;
			
			local moduleName = getXMLString(xmlFile, key .. "#name");
			if moduleName ~= nil and moduleName ~= "" then
				if not savedGame then
					GameExtension.addModule(moduleName, Utils.getNoNil(getXMLBool(xmlFile, key .. "#isSpec"), false));
				else
					deactivateOnMP = GameExtension.getSetting(moduleName, "DEACTIVATE_ON_MP"); -- If true dont load these settings.
					
					if isMultiplayer and deactivateOnMP then
						GameExtension.addBlackListItem(moduleName); -- black list so that this module state arent saved for client.
					end;
				end;
				
				-- Only do on server
				if isServer then
					if not Utils.getNoNil(getXMLBool(xmlFile, key .. "#active"), true) then
						-- All settings arent loaded on 1th load. Force on Call 1
						GameExtension.deactivateModule(moduleName, GameExtension.modName, not savedGame);
					end;
				end;
				
				if isMultiplayer and not deactivateOnMP or not isMultiplayer then
					local subI = 0;
					while true do
						local subKey = string.format(key .. ".setting(%d)", subI);
						if not hasXMLProperty(xmlFile, subKey) then break; end;
				
						local name = getXMLString(xmlFile, subKey .. "#name");
						local valueType = getXMLString(xmlFile, subKey .. "#type");
						local value, entry, minMax;
						
						if valueType == "string" then
							value = getXMLString(xmlFile, subKey .. "#value");
						elseif valueType == "float" then
							value = getXMLFloat(xmlFile, subKey .. "#value");
						elseif valueType == "int" then
							value = getXMLInt(xmlFile, subKey .. "#value");
						elseif valueType == "bool" then
							value = getXMLBool(xmlFile, subKey .. "#value");
						end;
						
						if value ~= nil then
							if not savedGame then
								local limits = getXMLString(xmlFile, subKey .. "#limits");
								if limits ~= nil and limits ~= "" then
									local values = Utils.getVectorNFromString(limits)
									minMax = {min = values[1], max = values[2]};
								end;
								
								entry = {};
								entry.type = valueType;
								entry.limits = minMax;
								entry.value = value;
								entry.showInSettings = Utils.getNoNil(getXMLBool(xmlFile, subKey .. "#showInSettings"), true);
								entry.isLocked = false; 		-- Dont change this
								entry.isLockedByForce = false; 	-- This can be changed by modders
								entry.func = getXMLString(xmlFile, subKey .. "#func");
								entry.allowEvent = Utils.getNoNil(getXMLBool(xmlFile, subKey .. "#event"), true);
								
								if Utils.getNoNil(getXMLBool(xmlFile, subKey .. "#blackListGUI"), false) then
									GameExtension.addBlackListItem(name);
								end;
								
								GameExtension.setSetting(moduleName, name, entry, true, true);
							else
								GameExtension.setSetting(moduleName, name, value, true, false);
							end;
						end;
						
						subI = subI + 1;
					end;
				end;
			end;
			
			i = i + 1;
		end;
		
		delete(xmlFile);
	else
		log("DEBUG", "Cant load xmlFile: " .. xmlPath);
	end;
end;



-- Misc --

function GameExtension:getIsActiveForInput()
	if g_gui:getIsGuiVisible() or g_currentMission.isPlayerFrozen or g_currentMission.inGameMessage:getIsVisible() then
		return false;
	end;
	
	return true;
end;

function GameExtension.makeTextGlobal(name)
	if g_i18n:hasText(name) then
		g_i18n.globalI18N.texts[name] = g_i18n:getText(name);
	else
		log("ERROR", "Can't make text ( " .. tostring(name) .. " ) global as we can't find it!");
	end;
end;

-- {bottom left X, bottom left Y, top right X, top right Y}, {file size width, file size height}
function GameExtension.normalizeUVs(uv, ref)
	local uvs = {
		uv[1], uv[2],	-- v0
		uv[1], uv[4],	-- v1
		uv[3], uv[2],	-- v2
		uv[3], uv[4]	-- v3
	};
	
	uvs = getNormalizedValues(uvs, Utils.getNoNil(ref, {512, 512}));
	
	return uvs;
end;

function GameExtension.inverseValue(value, offset)
	offset = Utils.getNoNil(offset, 0);
	
	return (1 - value) - offset;
end;

function GameExtension.loadHudIcon(name, uv, color)
	local uiScale 	= 0.6 * GameExtension.getSetting("VEHICLE", "VEHICLE_HUD_SCALE") * g_gameSettings:getValue("uiScale");
	local size 		= {getNormalizedScreenValues(47 * uiScale, 32 * uiScale)};
	local overlay = Overlay:new(name, GameExtension.hudElements, 0.5, 0.5, size[1], size[2]);
	
	overlay:setUVs(GameExtension.normalizeUVs(uv));
	
	if color ~= nil then
		overlay:setColor(unpack(GameExtension.colorCodes[color]));
	end;
	
	return overlay;
end;

function GameExtension:callFunc(name, values)
	for moduleName, v in pairs(GameExtension.modules) do
		if v.class ~= nil and v.class[name] ~= nil and not v.isSpec then
			v.class[name](self, unpack(values));
		end;
	end;
end;

function GameExtension.callSpecFunc(vehicle, name, values)
	for i, v in pairs(GameExtension.specializations) do
		if v[name] ~= nil then
			v[name](vehicle, unpack(values));
		end;
	end;
end;

function GameExtension.addClassOverride(name, class, func, new)
	if GameExtension.classOverrides[name] == nil then
		GameExtension.classOverrides[name] = {class = class, func = func, new = new};
	end;
end;

function GameExtension.overrideModClass(modName, className, target, new)
	if g_modIsLoaded[modName] then
		local class = _G[modName][className];
		class[target] = Utils.overwrittenFunction(class[target], new);
	end;
end;

function GameExtension.disableModClass(modName, className)
	if g_modIsLoaded[modName] then
		for i, v in ipairs(g_modEventListeners) do
			if v == _G[modName][className] then
				table.remove(g_modEventListeners, i);
				break;
			end;
		end;
	end;
end;



function GameExtension.getFruitArea(fruitId, startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, allowPreparing, useMinForageState)
	if allowPreparing == nil then
		allowPreparing = false
	end

	local ids = g_currentMission.fruits[fruitId]

	if ids == nil or ids.id == 0 then
		return 0, 0, 0
	end

	local id = ids.id
	local desc = FruitUtil.fruitIndexToDesc[fruitId]
	local x, z, widthX, widthZ, heightX, heightZ = Utils.getXZWidthAndHeight(id, startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ);
	local density, numberOfPixels, totalNumberOfPixels = getDensityParallelogram(id, x, z, widthX, widthZ, heightX, heightZ, 0, g_currentMission.numFruitStateChannels);
	local growthState, biggestFound = 0, 0;
	
	growthState = density / numberOfPixels; -- average, reducing the size where we look to very small reduce the chance of being between 2 states and getting the wrong growthState as an result.
	
	-- Fetch the biggest density value, will always get the correct growthState but at the cost of performance!
	-- for i = 0, 10 do -- do we need to do all 15?
		-- setDensityCompareParams(id, "equal", i + 1);
		-- local d, n, p = getDensityParallelogram(id, x, z, widthX, widthZ, heightX, heightZ, 0, g_currentMission.numFruitStateChannels);
	
		-- if d > biggestFound then
			-- growthState = i + 1;
			-- biggestFound = d;
			-- density = d;
			-- numberOfPixels = n
			-- totalNumberOfPixels = p;
		-- end;
	-- end;
	
	-- setDensityCompareParams(id, "greater", -1); -- reset setDensityCompareParams
	
	return density, numberOfPixels, growthState, totalNumberOfPixels;
end;