--
-- GameExtensionGUI
--
-- @author:    	Xentro (Marcus@Xentro.se)
-- @website:	www.Xentro.se
-- @history		v1.0 - 2016-11-10 - Initial implementation
-- 

GameExtensionGUI = {};

-- Page marker
GameExtensionGUI.NORMAL = 1;
GameExtensionGUI.FOCUS = 3;

GameExtensionGUI.SETTINGS_PER_LINE = 6;
GameExtensionGUI.SETTINGS_PER_PAGE = GameExtensionGUI.SETTINGS_PER_LINE * 3;

local GameExtensionGUI_mt = Class(GameExtensionGUI, ScreenElement);

function GameExtensionGUI:new(target, custom_mt)
	if custom_mt == nil then
		custom_mt = GameExtensionGUI_mt
	end;
	
	local self = ScreenElement:new(target, custom_mt);
	
	self.isOpen = false;
	
	self.clientSettings = {};
	self.serverSettings = {};
	
	self.settingValuesToRowIndex = {};
	self.settingRowIndexToValue = {};
	
	self.currentPage = 1;
	self.pages = {};
	self.customPages = {};
	self.isAdminPage = {};
	self.clientArentLoggedIn = nil;
	
	self.pageMarkers = {}
	
	return self;
end;

function GameExtensionGUI:addPage(name, settings)
	name = name:lower();
	
	if name == "client" or name == "server" then
		log(2, "Your not allowed to add this page ( " .. name .. " )");
	else
		if self.customPages[name] == nil and settings ~= nil then
			self.customPages[name] = settings;
		end;
	end;
end;

function GameExtensionGUI:addSettingsToPage(name, settings)
	name = name:lower();
	
	if name == "client" or name == "server" then
		-- TODO, we need to seperate them from gameExtensions settings..
		log(2, "Your not allowed to add settings to default client / server page.");
	else
		if self.customPages[name] ~= nil then 
			for _, v in ipairs(settings) do
				table.insert(self.customPages[name].settings, v);
			end;
		end;
	end;
end;

function GameExtensionGUI:load()
	function pairsByKeys(t, f)
		local a = {};
		
		for n in pairs(t) do table.insert(a, n) end
			table.sort(a, f);
			local i = 0; 
			local iter = function()
			i = i + 1;
			
			if a[i] == nil then
				return nil;
			else 
				return a[i], t[a[i]];
			end;
		end;
		
		return iter;
	end;
	
	for moduleName, s in pairsByKeys(GameExtension.modules) do
		if not GameExtension.getBlackListItem(moduleName) then
			local settings = GameExtension.getSetting(moduleName, nil, true);
			
			for i, name in ipairs(s.sortedKeys) do
				if not GameExtension.getBlackListItem(name) then
					local v = settings[name];
					
					if v.allowEvent then
						table.insert(self.serverSettings, {moduleName, name});
					else
						table.insert(self.clientSettings, {moduleName, name});
					end;
				end;
			end;
		end;
	end;
	
	self.settingItemTemplate:setVisible(false); -- Set to false as apparently it gets harder to do it later.
	self.pageMarkerTemplate:setVisible(false);
	
	self:setupPage(self.clientSettings, false, 0);
	self:setupPage(self.serverSettings, false, #self.pages);
	
	for name, v in pairs(self.customPages) do
		local temp = {};
		
		for _, v1 in ipairs(v.settings) do
			table.insert(temp, {name, v1});
		end;
		
		self:setupPage(temp, true, #self.pages);
	end;
	
	for i = 1, #self.pages, 1 do
		local marker = self.pageMarkerTemplate:clone(self.pageMarkerParent);
		marker:setVisible(true);
		self.pageMarkers[i] = marker;
	end;
	
	self:centerElements(self.pageMarkers, self.pageMarkers[1].size[1] * 3);
end;

function GameExtensionGUI:delete()
	-- Do we need to delete cloned elements ?
end;

function GameExtensionGUI:keyEvent(unicode, sym, modifier, isDown)
end;

function GameExtensionGUI:update(dt)
	if not g_gui:getIsDialogVisible("PasswordDialog") then
		if InputBinding.hasEvent(InputBinding.TOGGLE_GUI_SCREEN, true) or InputBinding.hasEvent(InputBinding.MENU_CANCEL, true) then
			self:exitGUI();
		end;
	end;
	
	if g_currentMission.missionDynamicInfo.isMultiplayer then
		-- Set the correct page if your logged in.
		if g_currentMission.isMasterUser then
			if self.clientArentLoggedIn == nil then -- delay 1 frame
				if g_gui_CurrentDialog ~= nil then
					-- Prevent dialog to open the ingameMenu
					g_gui_CurrentDialog.target:setCallback(GameExtensionGUI.onOK, self);
					g_gui_CurrentDialog.target:setReturnScreen(nil, nil);
					g_gui_CurrentDialog = nil;
				end;
			end;
			
			if self.clientArentLoggedIn ~= nil then
				self.clientArentLoggedIn = nil;
				self:setPage();
			end;
		end;
	end;
end;

function GameExtensionGUI:draw()
end;

function GameExtensionGUI:onOpen()
	GameExtensionGUI:superClass().onOpen(self);
	
	g_currentMission.GameExtension_GUI_isOpen = true;
	g_currentMission.isPlayerFrozen = true;
	
	self.isOpen = true;
	-- self.currentPage = 1;
	self:updateAllPages(); -- Update settings
	
	
	-- setup setting values, rework
	local pages = {};
	pages = self:loadPages(pages, self.clientSettings, "CLIENT_SETTING_PAGE", false);
	pages = self:loadPages(pages, self.serverSettings, "SERVER_SETTING_PAGE", true);
	
	for name, v in pairs(self.customPages) do
		pages = self:loadPages(pages, v.settings, v.pageName, v.isAdminPage);
	end;
	
	self.pageSelectionElement:setTexts(pages);
	self.pageSelectionElement:setState(self.currentPage);
	
	self:setPage();
end;

function GameExtensionGUI:loadPages(pages, settings, name, isAdminPage)
	local numNeededPages = math.ceil(#settings / GameExtensionGUI.SETTINGS_PER_PAGE);
	
	for i = 1, numNeededPages do
		if g_i18n:hasText(name) then
			name = string.format(g_i18n:getText(name), i);
		end;
		
		table.insert(pages, name);
		table.insert(self.isAdminPage, isAdminPage);
	end;
	
	return pages;
end;

function GameExtensionGUI:onClose()
	GameExtensionGUI:superClass().onClose(self);
	
	self.isOpen = false;
	g_currentMission.GameExtension_GUI_isOpen = false;
	g_currentMission.isPlayerFrozen = false;
end;

function GameExtensionGUI:setupPage(tableSetting, customPage, currentPage)
	local numEntrys = 0;
	local focusId;
	local currentPageElement = self.pageTemplate;
	local currentPage2Element = self.settingFrameTemplate;
	
	for i = 1, #tableSetting do
		local setting = tableSetting[i];
		local settings, value, name;
		
		if not customPage then
			settings = GameExtension.getSetting(setting[1], nil, true);
			settings = settings[setting[2]];
		else
			settings = setting[2];
		end;
		
		if settings.showInSettings then
			numEntrys = numEntrys + 1;
			
			if numEntrys == 1 or numEntrys > GameExtensionGUI.SETTINGS_PER_PAGE then
				currentPageElement = self.pageTemplate:clone(self.mainPageElement);
				table.insert(self.pages, {currentPageElement, {}});
				numEntrys = 1;
				focusId = 600;
			end;
		
			if numEntrys == (GameExtensionGUI.SETTINGS_PER_LINE * 2) + 1 then
				currentPage2Element = self.settingFrameTemplate:clone(currentPageElement);
				currentPage2Element:move(0.5, 0);
			elseif numEntrys == GameExtensionGUI.SETTINGS_PER_LINE + 1 then
				currentPage2Element = self.settingFrameTemplate:clone(currentPageElement);
				currentPage2Element:move(0.25, 0);
			elseif numEntrys == 1 then
				currentPage2Element = self.settingFrameTemplate:clone(currentPageElement);
				-- currentPage2Element:move(0, 0);
			end;
			
			local c = self.settingItemTemplate:clone(currentPage2Element);
			if numEntrys == GameExtensionGUI.SETTINGS_PER_LINE 
			or numEntrys == GameExtensionGUI.SETTINGS_PER_LINE * 2 
			or numEntrys == GameExtensionGUI.SETTINGS_PER_LINE * 3 then
				c:move(0, -0.52); -- Position reset for last line entry for some reason.. quick fix for now.
			end;
			
			if not customPage then
				value = GameExtension.getSetting(setting[1], setting[2]);
				name = setting[2];
				
				c.name = settings.type .. "." .. setting[1] .. "." .. setting[2];
				c.isCustomPage = false;
				c.toolTip = g_i18n:getText("toolTip_" .. name);
				c.elements[4].text = g_i18n:getText(name);
			else
				value = settings.parent[settings.variableName];
				name = settings.name;
				
				c.name = settings.type .. "." .. setting[1] .. "." .. settings.name;
				c.isCustomPage = true;
				c.settingId = i;
				c.toolTip = settings.toolTip;
				c.elements[4].text = settings.name;
			end;
			
			-- log("DEBUG", "element: " .. tostring(c) .. ", Parent1: " .. tostring(currentPage2Element) .. ", Parent2: " .. tostring(currentPageElement) .. " - name: " .. name);
			table.insert(self.pages[#self.pages][2], c);
			
			-- Setup values and set them
			self:setSettingElement(c, name, value, settings);
			
			c:setVisible(true);
			c.focusId = focusId;
			
			if c.focusChangeData ~= nil then
				table.insert(c.focusChangeData, focusId - 1);
				table.insert(c.focusChangeData, focusId + 1);
			end;
			
			focusId = focusId + 1;
			-- log("DEBUG", "Creating setting entry: " .. numEntrys .. ", " .. name);
		end;
	end;
	
	-- Quick fix for the last entry, position is reset.
	self.settingItemTemplate:clone(currentPage2Element);
end;

function GameExtensionGUI:updateAllPages()
	for _, v1 in ipairs(self.pages) do
		for _, v2 in ipairs(v1[2]) do
			self:setSettingElement(v2, nil);
		end;
	end;
end;

function GameExtensionGUI:setSettingElement(c, name, value, settings)
	local valueType = "bool";
	
	-- Update settings
	if name == nil then
		local split = Utils.splitString(".", c.name);
		valueType = split[1];
		name = split[3];
		
		if not c.isCustomPage then
			local allModuleSettings = GameExtension.getSetting(split[2], split[3], true);
			settings = allModuleSettings[name];
			
			if settings ~= nil then
				value = settings.value;
			end;
		else
			settings = self.customPages[split[2]].settings[c.settingId];
			value = settings.parent[settings.variableName];
		end;
	end;
	
	-- Is empty for module state
	if settings ~= nil then
		valueType = settings.type;
	end;
	
	if valueType == "string" then
		-- not supported
	elseif valueType == "float" or valueType == "int" then
		local row = 1;
		
		if self.settingValuesToRowIndex[name] == nil then
			local range = settings.limits.max - settings.limits.min;
			local min = settings.limits.min;
			
			self.settingValuesToRowIndex[name] = {};
			self.settingRowIndexToValue[name] = {};
			
			for k = 1, range + 1 do
				if valueType == "int" then
					local v = k + min - 1;
					self.settingValuesToRowIndex[name][v] = k;
					table.insert(self.settingRowIndexToValue[name], tostring(v));
				else
					-- TODO, Remake
					if k <= range then -- avoid the last 
						for l = 1, 5 do
							local v = k + (0.25 * (l - 1)) + min - 1;
							
							if self.settingValuesToRowIndex[name][v] == nil then
								table.insert(self.settingRowIndexToValue[name], tostring(v));
								self.settingValuesToRowIndex[name][v] = #self.settingRowIndexToValue[name];
							end;
						end;
					end;
				end;
			end;
		end;
		
		row = self.settingValuesToRowIndex[name][value];
		c:setTexts(self.settingRowIndexToValue[name]);
		c:setState(row);
	else
		local page = 1;
		if value then
			page = 2;
		end;
		
		c:setState(page);
	end;
	
	c:setDisabled(settings.isLocked or settings.isLockedByForce);
end;

function GameExtensionGUI:setSetting(value, element)
	local res = Utils.splitString(".", element.name);
	
	if res[1] == "string" then
		-- 
	elseif res[1] == "float" or res[1] == "int" then
		value = tonumber(self.settingRowIndexToValue[res[3]][value]);
	elseif res[1] == "bool" then
		value = value == 2; -- convert to bool
	end;
	
	if not element.isCustomPage then
		GameExtension.setSetting(res[2], res[3], value);
	else
		settings = self.customPages[res[2]].settings[element.settingId];
		
		if settings.func ~= nil then
			settings.func(settings.parent, value);
		else
			-- settings.parent[settings.variableName] = value;
			log(2, "This is currently not supported.");
		end;
	end;
end;

function GameExtensionGUI:setPage()
	local showPage = true;
	
	-- Handle redirection to login "page"
	if self.isAdminPage[self.currentPage] then
		if g_currentMission.missionDynamicInfo.isMultiplayer and not g_currentMission:getIsServer() then
			if not g_currentMission.isMasterUser then
				showPage = false;
				self.clientArentLoggedIn = true;
			end;
		end;
	end;
	
	if showPage then
		for i, element in ipairs(self.pages) do
			element[1]:setVisible(i == self.currentPage);
		end;
		
		self.loginElement:setVisible(false);
	else
		for i, element in ipairs(self.pages) do
			element[1]:setVisible(false);
		end;
		
		self.loginElement:setVisible(true);
	end;
	
	for i, marker in ipairs(self.pageMarkers) do
		if i == self.currentPage then
			marker.state = GameExtensionGUI.FOCUS;
		else
			marker.state = GameExtensionGUI.NORMAL;
		end;
	end;
end;


function GameExtensionGUI:createMainPage(element)
	self.mainPageElement = element;
end;


-- Selection --
function GameExtensionGUI:createPageSelection(element)
	self.pageSelectionElement = element;
end;

function GameExtensionGUI:updatePageSelection(row)
	self.currentPage = row;
	
	self:setPage();
end;


-- page element
function GameExtensionGUI:createPage(element)
	if self.pageTemplate == nil then
		self.pageTemplate = element;
	end;
end;

function GameExtensionGUI:createSettingFrame(element)
	if self.settingFrameTemplate == nil then
		self.settingFrameTemplate = element;
	end;
end;
function GameExtensionGUI:createSettingItem(element)
	if self.settingItemTemplate == nil then
		self.settingItemTemplate = element;
	end;
end;

function GameExtensionGUI:onFocusEnterSettingItem(element)
	self.helpBoxElement:setVisible(false);
end;

function GameExtensionGUI:onFocusLeaveSettingItem(element)
	if element.toolTip ~= nil then
		self.helpBoxElement:setVisible(true);
		self.helpBoxElement.elements[2]:setText(element.toolTip);
	end;
end;

-- page login
function GameExtensionGUI:createLogin(element)
	self.loginElement = element;
end;

function GameExtensionGUI:onClickLogin()
	-- if g_dedicatedServerInfo == nil and g_currentMission.missionDynamicInfo.isMultiplayer and not g_currentMission:getIsServer() then
		-- We could open the error dialog here instead of opening the password dialog
	-- else
		local dialog = g_gui:showDialog("PasswordDialog");
		dialog.target:setCallback(GameExtensionGUI.passwordCallback, self);
	-- end;
end;

function GameExtensionGUI:passwordCallback(password, login)
	if GameExtension.HIDDEN_LOGIN_PASSWORD then
		if g_passwordDialog.passwordElement.realText ~= nil then
			password = g_passwordDialog.passwordElement.realText;
		end;
	end;
	
	if login then
		g_client:getServerConnection():sendEvent(GetAdminEvent:new(password));
	else
		g_gui:closeDialogByName("PasswordDialog");
	end;
end;


-- help box
function GameExtensionGUI:createPageMarkerBox(element)
	self.pageMarkerParent = element;
end;

function GameExtensionGUI:createHelpBox(element)
	self.helpBoxElement = element;
end;


-- 
function GameExtensionGUI:createPageMarker(element)
	if self.pageMarkerTemplate == nil then
		self.pageMarkerTemplate = element;
	end;
end;


-- Button
function GameExtensionGUI:onClickBack()
	self:setReturnScreen("", "");
	
	GameExtensionGUI:superClass().onClickBack(self);
end;


-- Exit GUI
function GameExtensionGUI:exitGUI()
	g_gui:showGui("");
end;


-- Dialog fix
local oldFunc = Gui.showDialog;
Gui.showDialog = function(self, name)
	local dialog = oldFunc(self, name);
	
	if name == "InfoDialog" and g_currentMission.GameExtension_GUI_isOpen then
		g_gui_CurrentDialog = dialog;
	end;
	
	return dialog;
end;

function GameExtensionGUI.onOK(self)
	g_gui:closeAllDialogs();
end;